import { __assign, __rest } from "tslib";
import { useRef, useEffect, useCallback, useMemo, useState, } from 'react';
import { equal } from '@wry/equality';
import { ApolloError, } from "../../core/index.js";
import { invariant } from "../../utilities/globals/index.js";
import { compact, isNonEmptyArray, hasDirectives, } from "../../utilities/index.js";
import { useApolloClient } from "./useApolloClient.js";
import { DocumentType, verifyDocumentType } from "../parser/index.js";
import { useDeepMemo, useIsomorphicLayoutEffect } from "./internal/index.js";
import { useSuspenseCache } from "./useSuspenseCache.js";
import { useSyncExternalStore } from "./useSyncExternalStore.js";
var SUPPORTED_FETCH_POLICIES = [
    'cache-first',
    'network-only',
    'no-cache',
    'cache-and-network',
];
var DEFAULT_FETCH_POLICY = 'cache-first';
var DEFAULT_SUSPENSE_POLICY = 'always';
var DEFAULT_ERROR_POLICY = 'none';
export function useSuspenseQuery_experimental(query, options) {
    if (options === void 0) { options = Object.create(null); }
    var suspenseCache = useSuspenseCache();
    var client = useApolloClient(options.client);
    var watchQueryOptions = useWatchQueryOptions({ query: query, options: options, client: client });
    var previousWatchQueryOptionsRef = useRef(watchQueryOptions);
    var deferred = useIsDeferred(query);
    var fetchPolicy = watchQueryOptions.fetchPolicy, errorPolicy = watchQueryOptions.errorPolicy, returnPartialData = watchQueryOptions.returnPartialData, variables = watchQueryOptions.variables;
    var cacheEntry = suspenseCache.lookup(query, variables);
    var observable = useState(function () {
        return (cacheEntry === null || cacheEntry === void 0 ? void 0 : cacheEntry.observable) || client.watchQuery(watchQueryOptions);
    })[0];
    var result = useObservableQueryResult(observable);
    var hasFullResult = result.data && !result.partial;
    var hasPartialResult = result.data && result.partial;
    var usePartialResult = returnPartialData && hasPartialResult;
    var allowsThrownErrors = errorPolicy === 'none' && (!deferred || !hasPartialResult);
    if (result.error &&
        (result.error.networkError || allowsThrownErrors)) {
        throw result.error;
    }
    if (result.loading) {
        if (!cacheEntry) {
            cacheEntry = suspenseCache.add(query, variables, {
                promise: maybeWrapConcastWithCustomPromise(observable.reobserveAsConcast(watchQueryOptions), { deferred: deferred }),
                observable: observable,
            });
        }
        var hasUsableResult = usePartialResult ||
            (fetchPolicy === 'cache-and-network' && hasFullResult);
        if (!hasUsableResult && !cacheEntry.fulfilled) {
            throw cacheEntry.promise;
        }
    }
    useEffect(function () {
        var variables = watchQueryOptions.variables, query = watchQueryOptions.query;
        var previousOpts = previousWatchQueryOptionsRef.current;
        if (variables !== previousOpts.variables || query !== previousOpts.query) {
            suspenseCache.remove(previousOpts.query, previousOpts.variables);
            suspenseCache.add(query, variables, {
                promise: observable.reobserve({ query: query, variables: variables }),
                observable: observable,
            });
            previousWatchQueryOptionsRef.current = watchQueryOptions;
        }
    }, [watchQueryOptions]);
    useEffect(function () {
        return function () {
            suspenseCache.remove(query, variables);
        };
    }, []);
    return useMemo(function () {
        return {
            data: result.data,
            error: errorPolicy === 'ignore' ? void 0 : toApolloError(result),
            fetchMore: function (options) {
                var promise = observable.fetchMore(options);
                suspenseCache.add(query, watchQueryOptions.variables, {
                    promise: promise,
                    observable: observable,
                });
                return promise;
            },
            refetch: function (variables) {
                var promise = observable.refetch(variables);
                suspenseCache.add(query, watchQueryOptions.variables, {
                    promise: promise,
                    observable: observable,
                });
                return promise;
            },
        };
    }, [result, observable, errorPolicy]);
}
function validateOptions(options) {
    var query = options.query, _a = options.fetchPolicy, fetchPolicy = _a === void 0 ? DEFAULT_FETCH_POLICY : _a, returnPartialData = options.returnPartialData;
    verifyDocumentType(query, DocumentType.Query);
    validateFetchPolicy(fetchPolicy);
    validatePartialDataReturn(fetchPolicy, returnPartialData);
}
function validateFetchPolicy(fetchPolicy) {
    __DEV__ ? invariant(SUPPORTED_FETCH_POLICIES.includes(fetchPolicy), "The fetch policy `".concat(fetchPolicy, "` is not supported with suspense.")) : invariant(SUPPORTED_FETCH_POLICIES.includes(fetchPolicy), 33);
}
function validatePartialDataReturn(fetchPolicy, returnPartialData) {
    if (fetchPolicy === 'no-cache' && returnPartialData) {
        __DEV__ && invariant.warn('Using `returnPartialData` with a `no-cache` fetch policy has no effect. To read partial data from the cache, consider using an alternate fetch policy.');
    }
}
function toApolloError(result) {
    return isNonEmptyArray(result.errors)
        ? new ApolloError({ graphQLErrors: result.errors })
        : result.error;
}
function maybeWrapConcastWithCustomPromise(concast, _a) {
    var deferred = _a.deferred;
    if (deferred) {
        return new Promise(function (resolve, reject) {
            var subscription = concast.subscribe({
                next: function (value) {
                    resolve(value);
                    subscription.unsubscribe();
                },
                error: reject,
            });
        });
    }
    return concast.promise;
}
function useWatchQueryOptions(_a) {
    var query = _a.query, options = _a.options, client = _a.client;
    var defaultOptions = client.defaultOptions.watchQuery;
    var watchQueryOptions = useDeepMemo(function () {
        var errorPolicy = options.errorPolicy, fetchPolicy = options.fetchPolicy, _a = options.suspensePolicy, suspensePolicy = _a === void 0 ? DEFAULT_SUSPENSE_POLICY : _a, variables = options.variables, watchQueryOptions = __rest(options, ["errorPolicy", "fetchPolicy", "suspensePolicy", "variables"]);
        return __assign(__assign({}, watchQueryOptions), { query: query, errorPolicy: errorPolicy || (defaultOptions === null || defaultOptions === void 0 ? void 0 : defaultOptions.errorPolicy) || DEFAULT_ERROR_POLICY, fetchPolicy: fetchPolicy || (defaultOptions === null || defaultOptions === void 0 ? void 0 : defaultOptions.fetchPolicy) || DEFAULT_FETCH_POLICY, notifyOnNetworkStatusChange: suspensePolicy === 'always', fetchOnFirstSubscribe: false, variables: compact(__assign(__assign({}, defaultOptions === null || defaultOptions === void 0 ? void 0 : defaultOptions.variables), variables)) });
    }, [options, query, defaultOptions]);
    if (__DEV__) {
        validateOptions(watchQueryOptions);
    }
    return watchQueryOptions;
}
function useIsDeferred(query) {
    return useMemo(function () { return hasDirectives(['defer'], query); }, [query]);
}
function useObservableQueryResult(observable) {
    var resultRef = useRef();
    var isMountedRef = useRef(false);
    if (!resultRef.current) {
        resultRef.current = observable.getCurrentResult();
    }
    useIsomorphicLayoutEffect(function () {
        isMountedRef.current = true;
        return function () {
            isMountedRef.current = false;
        };
    }, []);
    return useSyncExternalStore(useCallback(function (forceUpdate) {
        function handleUpdate() {
            var previousResult = resultRef.current;
            var result = observable.getCurrentResult();
            if (previousResult.loading === result.loading &&
                previousResult.networkStatus === result.networkStatus &&
                equal(previousResult.data, result.data)) {
                return;
            }
            resultRef.current = result;
            if (isMountedRef.current) {
                forceUpdate();
            }
        }
        var subscription = observable.subscribe({
            next: handleUpdate,
            error: handleUpdate,
        });
        return function () {
            subscription.unsubscribe();
        };
    }, [observable]), function () { return resultRef.current; }, function () { return resultRef.current; });
}
//# sourceMappingURL=useSuspenseQuery.js.map